@@ -1,6 +1,6 @@
 /* ====================================================================
- * Copyright (c) 1999-2004 Ralf S. Engelschall <rse@engelschall.com>
- * Copyright (c) 1999-2004 The OSSP Project <http://www.ossp.org/>
+ * Copyright (c) 1999-2005 Ralf S. Engelschall <rse@engelschall.com>
+ * Copyright (c) 1999-2005 The OSSP Project <http://www.ossp.org/>
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -188,13 +188,17 @@ static void mm_insert_chunk(MM *mm, mem_chunk *mcInsert)
 {
     mem_chunk *mc;
     mem_chunk *mcPrev;
+    mem_chunk *mcPrevPrev;
     mem_chunk *mcNext;
 
     if (!mm_core_lock((void *)mm, MM_LOCK_RW))
         return;
     mc = &(mm->mp_freechunks);
-    while (mc->mc_u.mc_next != NULL && (char *)(mc->mc_u.mc_next) < (char *)mcInsert)
+    mcPrevPrev = mc;
+    while (mc->mc_u.mc_next != NULL && (char *)(mc->mc_u.mc_next) < (char *)mcInsert) {
+        mcPrevPrev = mc;
         mc = mc->mc_u.mc_next;
+    }
     mcPrev = mc;
     mcNext = mc->mc_u.mc_next;
     if (mcPrev == mcInsert || mcNext == mcInsert) {
@@ -209,6 +213,13 @@ static void mm_insert_chunk(MM *mm, mem_chunk *mcInsert)
         mcPrev->mc_u.mc_next = mcNext->mc_u.mc_next;
         mm->mp_freechunks.mc_usize -= 1;
     }
+    else if ((char *)mcPrev+(mcPrev->mc_size) == (char *)mcInsert &&
+             (char *)mcInsert+(mcInsert->mc_size) == ((char *)mm + mm->mp_offset)) {
+        /* merge with previous and spare block (to increase spare area) */
+        mcPrevPrev->mc_u.mc_next = mcPrev->mc_u.mc_next;
+        mm->mp_offset -= (mcInsert->mc_size + mcPrev->mc_size);
+        mm->mp_freechunks.mc_usize -= 1;
+    }
     else if ((char *)mcPrev+(mcPrev->mc_size) == (char *)mcInsert) {
         /* merge with previous chunk */
         mcPrev->mc_size += mcInsert->mc_size;
@@ -219,6 +230,10 @@ static void mm_insert_chunk(MM *mm, mem_chunk *mcInsert)
         mcInsert->mc_u.mc_next = mcNext->mc_u.mc_next;
         mcPrev->mc_u.mc_next = mcInsert;
     }
+    else if ((char *)mcInsert+(mcInsert->mc_size) == ((char *)mm + mm->mp_offset)) {
+        /* merge with spare block (to increase spare area) */
+        mm->mp_offset -= mcInsert->mc_size;
+    }
     else {
         /* no merging possible, so insert as new chunk */
         mcInsert->mc_u.mc_next = mcNext;
@@ -350,7 +365,7 @@ void *mm_realloc(MM *mm, void *ptr, size_t usize)
     }
     if ((vp = mm_malloc(mm, usize)) == NULL)
         return NULL;
-    memcpy(vp, ptr, usize);
+    memcpy(vp, ptr, mc->mc_usize);
     mm_free(mm, ptr);
     return vp;
 }
